# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_key import LedgerKey
from .revoke_sponsorship_op_signer import RevokeSponsorshipOpSigner
from .revoke_sponsorship_type import RevokeSponsorshipType

__all__ = ["RevokeSponsorshipOp"]


class RevokeSponsorshipOp:
    """
    XDR Source Code::

        union RevokeSponsorshipOp switch (RevokeSponsorshipType type)
        {
        case REVOKE_SPONSORSHIP_LEDGER_ENTRY:
            LedgerKey ledgerKey;
        case REVOKE_SPONSORSHIP_SIGNER:
            struct
            {
                AccountID accountID;
                SignerKey signerKey;
            } signer;
        };
    """

    def __init__(
        self,
        type: RevokeSponsorshipType,
        ledger_key: LedgerKey = None,
        signer: RevokeSponsorshipOpSigner = None,
    ) -> None:
        self.type = type
        self.ledger_key = ledger_key
        self.signer = signer

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY:
            if self.ledger_key is None:
                raise ValueError("ledger_key should not be None.")
            self.ledger_key.pack(packer)
            return
        if self.type == RevokeSponsorshipType.REVOKE_SPONSORSHIP_SIGNER:
            if self.signer is None:
                raise ValueError("signer should not be None.")
            self.signer.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> RevokeSponsorshipOp:
        type = RevokeSponsorshipType.unpack(unpacker)
        if type == RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY:
            ledger_key = LedgerKey.unpack(unpacker)
            return cls(type=type, ledger_key=ledger_key)
        if type == RevokeSponsorshipType.REVOKE_SPONSORSHIP_SIGNER:
            signer = RevokeSponsorshipOpSigner.unpack(unpacker)
            return cls(type=type, signer=signer)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> RevokeSponsorshipOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> RevokeSponsorshipOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.ledger_key,
                self.signer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.ledger_key == other.ledger_key
            and self.signer == other.signer
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"ledger_key={self.ledger_key}"
        ) if self.ledger_key is not None else None
        out.append(f"signer={self.signer}") if self.signer is not None else None
        return f"<RevokeSponsorshipOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .create_contract_args import CreateContractArgs
from .invoke_contract_args import InvokeContractArgs
from .soroban_authorized_function_type import SorobanAuthorizedFunctionType

__all__ = ["SorobanAuthorizedFunction"]


class SorobanAuthorizedFunction:
    """
    XDR Source Code::

        union SorobanAuthorizedFunction switch (SorobanAuthorizedFunctionType type)
        {
        case SOROBAN_AUTHORIZED_FUNCTION_TYPE_CONTRACT_FN:
            InvokeContractArgs contractFn;
        case SOROBAN_AUTHORIZED_FUNCTION_TYPE_CREATE_CONTRACT_HOST_FN:
            CreateContractArgs createContractHostFn;
        };
    """

    def __init__(
        self,
        type: SorobanAuthorizedFunctionType,
        contract_fn: InvokeContractArgs = None,
        create_contract_host_fn: CreateContractArgs = None,
    ) -> None:
        self.type = type
        self.contract_fn = contract_fn
        self.create_contract_host_fn = create_contract_host_fn

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if (
            self.type
            == SorobanAuthorizedFunctionType.SOROBAN_AUTHORIZED_FUNCTION_TYPE_CONTRACT_FN
        ):
            if self.contract_fn is None:
                raise ValueError("contract_fn should not be None.")
            self.contract_fn.pack(packer)
            return
        if (
            self.type
            == SorobanAuthorizedFunctionType.SOROBAN_AUTHORIZED_FUNCTION_TYPE_CREATE_CONTRACT_HOST_FN
        ):
            if self.create_contract_host_fn is None:
                raise ValueError("create_contract_host_fn should not be None.")
            self.create_contract_host_fn.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SorobanAuthorizedFunction:
        type = SorobanAuthorizedFunctionType.unpack(unpacker)
        if (
            type
            == SorobanAuthorizedFunctionType.SOROBAN_AUTHORIZED_FUNCTION_TYPE_CONTRACT_FN
        ):
            contract_fn = InvokeContractArgs.unpack(unpacker)
            return cls(type=type, contract_fn=contract_fn)
        if (
            type
            == SorobanAuthorizedFunctionType.SOROBAN_AUTHORIZED_FUNCTION_TYPE_CREATE_CONTRACT_HOST_FN
        ):
            create_contract_host_fn = CreateContractArgs.unpack(unpacker)
            return cls(type=type, create_contract_host_fn=create_contract_host_fn)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SorobanAuthorizedFunction:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SorobanAuthorizedFunction:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.contract_fn,
                self.create_contract_host_fn,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.contract_fn == other.contract_fn
            and self.create_contract_host_fn == other.create_contract_host_fn
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"contract_fn={self.contract_fn}"
        ) if self.contract_fn is not None else None
        out.append(
            f"create_contract_host_fn={self.create_contract_host_fn}"
        ) if self.create_contract_host_fn is not None else None
        return f"<SorobanAuthorizedFunction [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .contract_executable import ContractExecutable
from .sc_map import SCMap

__all__ = ["SCContractInstance"]


class SCContractInstance:
    """
    XDR Source Code::

        struct SCContractInstance {
            ContractExecutable executable;
            SCMap* storage;
        };
    """

    def __init__(
        self,
        executable: ContractExecutable,
        storage: Optional[SCMap],
    ) -> None:
        self.executable = executable
        self.storage = storage

    def pack(self, packer: Packer) -> None:
        self.executable.pack(packer)
        if self.storage is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.storage.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCContractInstance:
        executable = ContractExecutable.unpack(unpacker)
        storage = SCMap.unpack(unpacker) if unpacker.unpack_uint() else None
        return cls(
            executable=executable,
            storage=storage,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCContractInstance:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCContractInstance:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.executable,
                self.storage,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.executable == other.executable and self.storage == other.storage

    def __str__(self):
        out = [
            f"executable={self.executable}",
            f"storage={self.storage}",
        ]
        return f"<SCContractInstance [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .soroban_authorized_function import SorobanAuthorizedFunction

__all__ = ["SorobanAuthorizedInvocation"]


class SorobanAuthorizedInvocation:
    """
    XDR Source Code::

        struct SorobanAuthorizedInvocation
        {
            SorobanAuthorizedFunction function;
            SorobanAuthorizedInvocation subInvocations<>;
        };
    """

    def __init__(
        self,
        function: SorobanAuthorizedFunction,
        sub_invocations: List["SorobanAuthorizedInvocation"],
    ) -> None:
        _expect_max_length = 4294967295
        if sub_invocations and len(sub_invocations) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `sub_invocations` should be {_expect_max_length}, but got {len(sub_invocations)}."
            )
        self.function = function
        self.sub_invocations = sub_invocations

    def pack(self, packer: Packer) -> None:
        self.function.pack(packer)
        packer.pack_uint(len(self.sub_invocations))
        for sub_invocations_item in self.sub_invocations:
            sub_invocations_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SorobanAuthorizedInvocation:
        function = SorobanAuthorizedFunction.unpack(unpacker)
        length = unpacker.unpack_uint()
        sub_invocations = []
        for _ in range(length):
            sub_invocations.append(SorobanAuthorizedInvocation.unpack(unpacker))
        return cls(
            function=function,
            sub_invocations=sub_invocations,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SorobanAuthorizedInvocation:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SorobanAuthorizedInvocation:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.function,
                self.sub_invocations,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.function == other.function
            and self.sub_invocations == other.sub_invocations
        )

    def __str__(self):
        out = [
            f"function={self.function}",
            f"sub_invocations={self.sub_invocations}",
        ]
        return f"<SorobanAuthorizedInvocation [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_entry_data import LedgerEntryData
from .ledger_entry_ext import LedgerEntryExt
from .uint32 import Uint32

__all__ = ["LedgerEntry"]


class LedgerEntry:
    """
    XDR Source Code::

        struct LedgerEntry
        {
            uint32 lastModifiedLedgerSeq; // ledger the LedgerEntry was last changed

            union switch (LedgerEntryType type)
            {
            case ACCOUNT:
                AccountEntry account;
            case TRUSTLINE:
                TrustLineEntry trustLine;
            case OFFER:
                OfferEntry offer;
            case DATA:
                DataEntry data;
            case CLAIMABLE_BALANCE:
                ClaimableBalanceEntry claimableBalance;
            case LIQUIDITY_POOL:
                LiquidityPoolEntry liquidityPool;
            case CONTRACT_DATA:
                ContractDataEntry contractData;
            case CONTRACT_CODE:
                ContractCodeEntry contractCode;
            case CONFIG_SETTING:
                ConfigSettingEntry configSetting;
            case TTL:
                TTLEntry ttl;
            }
            data;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            case 1:
                LedgerEntryExtensionV1 v1;
            }
            ext;
        };
    """

    def __init__(
        self,
        last_modified_ledger_seq: Uint32,
        data: LedgerEntryData,
        ext: LedgerEntryExt,
    ) -> None:
        self.last_modified_ledger_seq = last_modified_ledger_seq
        self.data = data
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.last_modified_ledger_seq.pack(packer)
        self.data.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerEntry:
        last_modified_ledger_seq = Uint32.unpack(unpacker)
        data = LedgerEntryData.unpack(unpacker)
        ext = LedgerEntryExt.unpack(unpacker)
        return cls(
            last_modified_ledger_seq=last_modified_ledger_seq,
            data=data,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.last_modified_ledger_seq,
                self.data,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.last_modified_ledger_seq == other.last_modified_ledger_seq
            and self.data == other.data
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"last_modified_ledger_seq={self.last_modified_ledger_seq}",
            f"data={self.data}",
            f"ext={self.ext}",
        ]
        return f"<LedgerEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .fee_bump_transaction_ext import FeeBumpTransactionExt
from .fee_bump_transaction_inner_tx import FeeBumpTransactionInnerTx
from .int64 import Int64
from .muxed_account import MuxedAccount

__all__ = ["FeeBumpTransaction"]


class FeeBumpTransaction:
    """
    XDR Source Code::

        struct FeeBumpTransaction
        {
            MuxedAccount feeSource;
            int64 fee;
            union switch (EnvelopeType type)
            {
            case ENVELOPE_TYPE_TX:
                TransactionV1Envelope v1;
            }
            innerTx;
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        fee_source: MuxedAccount,
        fee: Int64,
        inner_tx: FeeBumpTransactionInnerTx,
        ext: FeeBumpTransactionExt,
    ) -> None:
        self.fee_source = fee_source
        self.fee = fee
        self.inner_tx = inner_tx
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.fee_source.pack(packer)
        self.fee.pack(packer)
        self.inner_tx.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> FeeBumpTransaction:
        fee_source = MuxedAccount.unpack(unpacker)
        fee = Int64.unpack(unpacker)
        inner_tx = FeeBumpTransactionInnerTx.unpack(unpacker)
        ext = FeeBumpTransactionExt.unpack(unpacker)
        return cls(
            fee_source=fee_source,
            fee=fee,
            inner_tx=inner_tx,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> FeeBumpTransaction:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> FeeBumpTransaction:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.fee_source,
                self.fee,
                self.inner_tx,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.fee_source == other.fee_source
            and self.fee == other.fee
            and self.inner_tx == other.inner_tx
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"fee_source={self.fee_source}",
            f"fee={self.fee}",
            f"inner_tx={self.inner_tx}",
            f"ext={self.ext}",
        ]
        return f"<FeeBumpTransaction [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .claimable_balance_id import ClaimableBalanceID
from .create_claimable_balance_result_code import CreateClaimableBalanceResultCode

__all__ = ["CreateClaimableBalanceResult"]


class CreateClaimableBalanceResult:
    """
    XDR Source Code::

        union CreateClaimableBalanceResult switch (
            CreateClaimableBalanceResultCode code)
        {
        case CREATE_CLAIMABLE_BALANCE_SUCCESS:
            ClaimableBalanceID balanceID;
        case CREATE_CLAIMABLE_BALANCE_MALFORMED:
        case CREATE_CLAIMABLE_BALANCE_LOW_RESERVE:
        case CREATE_CLAIMABLE_BALANCE_NO_TRUST:
        case CREATE_CLAIMABLE_BALANCE_NOT_AUTHORIZED:
        case CREATE_CLAIMABLE_BALANCE_UNDERFUNDED:
            void;
        };
    """

    def __init__(
        self,
        code: CreateClaimableBalanceResultCode,
        balance_id: ClaimableBalanceID = None,
    ) -> None:
        self.code = code
        self.balance_id = balance_id

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_SUCCESS
        ):
            if self.balance_id is None:
                raise ValueError("balance_id should not be None.")
            self.balance_id.pack(packer)
            return
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_MALFORMED
        ):
            return
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_LOW_RESERVE
        ):
            return
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_NO_TRUST
        ):
            return
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_NOT_AUTHORIZED
        ):
            return
        if (
            self.code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_UNDERFUNDED
        ):
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> CreateClaimableBalanceResult:
        code = CreateClaimableBalanceResultCode.unpack(unpacker)
        if code == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_SUCCESS:
            balance_id = ClaimableBalanceID.unpack(unpacker)
            return cls(code=code, balance_id=balance_id)
        if code == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_MALFORMED:
            return cls(code=code)
        if (
            code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_LOW_RESERVE
        ):
            return cls(code=code)
        if code == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_NO_TRUST:
            return cls(code=code)
        if (
            code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_NOT_AUTHORIZED
        ):
            return cls(code=code)
        if (
            code
            == CreateClaimableBalanceResultCode.CREATE_CLAIMABLE_BALANCE_UNDERFUNDED
        ):
            return cls(code=code)
        return cls(code=code)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> CreateClaimableBalanceResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> CreateClaimableBalanceResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.code,
                self.balance_id,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code and self.balance_id == other.balance_id

    def __str__(self):
        out = []
        out.append(f"code={self.code}")
        out.append(
            f"balance_id={self.balance_id}"
        ) if self.balance_id is not None else None
        return f"<CreateClaimableBalanceResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID

__all__ = ["SponsorshipDescriptor"]


class SponsorshipDescriptor:
    """
    XDR Source Code::

        typedef AccountID* SponsorshipDescriptor;
    """

    def __init__(self, sponsorship_descriptor: Optional[AccountID]) -> None:
        self.sponsorship_descriptor = sponsorship_descriptor

    def pack(self, packer: Packer) -> None:
        if self.sponsorship_descriptor is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.sponsorship_descriptor.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SponsorshipDescriptor:
        sponsorship_descriptor = (
            AccountID.unpack(unpacker) if unpacker.unpack_uint() else None
        )
        return cls(sponsorship_descriptor)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SponsorshipDescriptor:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SponsorshipDescriptor:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.sponsorship_descriptor)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.sponsorship_descriptor == other.sponsorship_descriptor

    def __str__(self):
        return f"<SponsorshipDescriptor [sponsorship_descriptor={self.sponsorship_descriptor}]>"



import json

__all__ = ["Response"]


class Response:
    """The :class:`Response <Response>` object, which contains a
    server's response to an HTTP request.

    :param status_code: response status code
    :param text: response content
    :param headers: response headers
    :param url: request url
    """

    def __init__(self, status_code: int, text: str, headers: dict, url: str) -> None:
        self.status_code: int = status_code
        self.text: str = text
        self.headers: dict = headers
        self.url: str = url

    def json(self) -> dict:
        """convert the content to dict

        :return: the content from server
        """
        return json.loads(self.text)

    def __hash__(self):
        return hash(
            (self.status_code, self.text, tuple(self.headers.items()), self.url)
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.status_code == other.status_code
            and self.text == other.text
            and self.headers == other.headers
            and self.url == other.url
        )

    def __str__(self):
        return (
            f"<Response [status_code={self.status_code}, "
            f"text={self.text}, "
            f"headers={self.headers}, "
            f"url={self.url}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .inner_transaction_result_ext import InnerTransactionResultExt
from .inner_transaction_result_result import InnerTransactionResultResult
from .int64 import Int64

__all__ = ["InnerTransactionResult"]


class InnerTransactionResult:
    """
    XDR Source Code::

        struct InnerTransactionResult
        {
            // Always 0. Here for binary compatibility.
            int64 feeCharged;

            union switch (TransactionResultCode code)
            {
            // txFEE_BUMP_INNER_SUCCESS is not included
            case txSUCCESS:
            case txFAILED:
                OperationResult results<>;
            case txTOO_EARLY:
            case txTOO_LATE:
            case txMISSING_OPERATION:
            case txBAD_SEQ:
            case txBAD_AUTH:
            case txINSUFFICIENT_BALANCE:
            case txNO_ACCOUNT:
            case txINSUFFICIENT_FEE:
            case txBAD_AUTH_EXTRA:
            case txINTERNAL_ERROR:
            case txNOT_SUPPORTED:
            // txFEE_BUMP_INNER_FAILED is not included
            case txBAD_SPONSORSHIP:
            case txBAD_MIN_SEQ_AGE_OR_GAP:
            case txMALFORMED:
            case txSOROBAN_INVALID:
                void;
            }
            result;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        fee_charged: Int64,
        result: InnerTransactionResultResult,
        ext: InnerTransactionResultExt,
    ) -> None:
        self.fee_charged = fee_charged
        self.result = result
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.fee_charged.pack(packer)
        self.result.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InnerTransactionResult:
        fee_charged = Int64.unpack(unpacker)
        result = InnerTransactionResultResult.unpack(unpacker)
        ext = InnerTransactionResultExt.unpack(unpacker)
        return cls(
            fee_charged=fee_charged,
            result=result,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InnerTransactionResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InnerTransactionResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.fee_charged,
                self.result,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.fee_charged == other.fee_charged
            and self.result == other.result
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"fee_charged={self.fee_charged}",
            f"result={self.result}",
            f"ext={self.ext}",
        ]
        return f"<InnerTransactionResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .liquidity_pool_entry_constant_product import LiquidityPoolEntryConstantProduct
from .liquidity_pool_type import LiquidityPoolType

__all__ = ["LiquidityPoolEntryBody"]


class LiquidityPoolEntryBody:
    """
    XDR Source Code::

        union switch (LiquidityPoolType type)
            {
            case LIQUIDITY_POOL_CONSTANT_PRODUCT:
                struct
                {
                    LiquidityPoolConstantProductParameters params;

                    int64 reserveA;        // amount of A in the pool
                    int64 reserveB;        // amount of B in the pool
                    int64 totalPoolShares; // total number of pool shares issued
                    int64 poolSharesTrustLineCount; // number of trust lines for the
                                                    // associated pool shares
                } constantProduct;
            }
    """

    def __init__(
        self,
        type: LiquidityPoolType,
        constant_product: LiquidityPoolEntryConstantProduct = None,
    ) -> None:
        self.type = type
        self.constant_product = constant_product

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == LiquidityPoolType.LIQUIDITY_POOL_CONSTANT_PRODUCT:
            if self.constant_product is None:
                raise ValueError("constant_product should not be None.")
            self.constant_product.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolEntryBody:
        type = LiquidityPoolType.unpack(unpacker)
        if type == LiquidityPoolType.LIQUIDITY_POOL_CONSTANT_PRODUCT:
            constant_product = LiquidityPoolEntryConstantProduct.unpack(unpacker)
            return cls(type=type, constant_product=constant_product)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolEntryBody:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolEntryBody:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.constant_product,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type and self.constant_product == other.constant_product
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"constant_product={self.constant_product}"
        ) if self.constant_product is not None else None
        return f"<LiquidityPoolEntryBody [{', '.join(out)}]>"



from typing import Union

from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseLedgersCallBuilder"]


class BaseLedgersCallBuilder(BaseCallBuilder):
    """Creates a new :class:`LedgersCallBuilder` pointed to server defined by horizon_url.

    See `List All Ledgers <https://developers.stellar.org/api/resources/ledgers/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "ledgers"

    def ledger(self, sequence: Union[int, str]):
        """Provides information on a single ledger.

        See `Retrieve a Ledger <https://developers.stellar.org/api/resources/ledgers/single/>`__ for more information.

        :param sequence: Ledger sequence
        :return: current LedgerCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}"
        return self



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .claimant_type import ClaimantType
from .claimant_v0 import ClaimantV0

__all__ = ["Claimant"]


class Claimant:
    """
    XDR Source Code::

        union Claimant switch (ClaimantType type)
        {
        case CLAIMANT_TYPE_V0:
            struct
            {
                AccountID destination;    // The account that can use this condition
                ClaimPredicate predicate; // Claimable if predicate is true
            } v0;
        };
    """

    def __init__(
        self,
        type: ClaimantType,
        v0: ClaimantV0 = None,
    ) -> None:
        self.type = type
        self.v0 = v0

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == ClaimantType.CLAIMANT_TYPE_V0:
            if self.v0 is None:
                raise ValueError("v0 should not be None.")
            self.v0.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Claimant:
        type = ClaimantType.unpack(unpacker)
        if type == ClaimantType.CLAIMANT_TYPE_V0:
            v0 = ClaimantV0.unpack(unpacker)
            return cls(type=type, v0=v0)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Claimant:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Claimant:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.v0,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.type == other.type and self.v0 == other.v0

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"v0={self.v0}") if self.v0 is not None else None
        return f"<Claimant [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque
from .extension_point import ExtensionPoint
from .hash import Hash

__all__ = ["ContractCodeEntry"]


class ContractCodeEntry:
    """
    XDR Source Code::

        struct ContractCodeEntry {
            ExtensionPoint ext;

            Hash hash;
            opaque code<>;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        hash: Hash,
        code: bytes,
    ) -> None:
        self.ext = ext
        self.hash = hash
        self.code = code

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.hash.pack(packer)
        Opaque(self.code, 4294967295, False).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractCodeEntry:
        ext = ExtensionPoint.unpack(unpacker)
        hash = Hash.unpack(unpacker)
        code = Opaque.unpack(unpacker, 4294967295, False)
        return cls(
            ext=ext,
            hash=hash,
            code=code,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractCodeEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractCodeEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.hash,
                self.code,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.hash == other.hash
            and self.code == other.code
        )

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"hash={self.hash}",
            f"code={self.code}",
        ]
        return f"<ContractCodeEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hash import Hash

__all__ = ["ConfigUpgradeSetKey"]


class ConfigUpgradeSetKey:
    """
    XDR Source Code::

        struct ConfigUpgradeSetKey {
            Hash contractID;
            Hash contentHash;
        };
    """

    def __init__(
        self,
        contract_id: Hash,
        content_hash: Hash,
    ) -> None:
        self.contract_id = contract_id
        self.content_hash = content_hash

    def pack(self, packer: Packer) -> None:
        self.contract_id.pack(packer)
        self.content_hash.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigUpgradeSetKey:
        contract_id = Hash.unpack(unpacker)
        content_hash = Hash.unpack(unpacker)
        return cls(
            contract_id=contract_id,
            content_hash=content_hash,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigUpgradeSetKey:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigUpgradeSetKey:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.contract_id,
                self.content_hash,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.contract_id == other.contract_id
            and self.content_hash == other.content_hash
        )

    def __str__(self):
        out = [
            f"contract_id={self.contract_id}",
            f"content_hash={self.content_hash}",
        ]
        return f"<ConfigUpgradeSetKey [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque

__all__ = ["UpgradeType"]


class UpgradeType:
    """
    XDR Source Code::

        typedef opaque UpgradeType<128>;
    """

    def __init__(self, upgrade_type: bytes) -> None:
        self.upgrade_type = upgrade_type

    def pack(self, packer: Packer) -> None:
        Opaque(self.upgrade_type, 128, False).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> UpgradeType:
        upgrade_type = Opaque.unpack(unpacker, 128, False)
        return cls(upgrade_type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> UpgradeType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> UpgradeType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.upgrade_type)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.upgrade_type == other.upgrade_type

    def __str__(self):
        return f"<UpgradeType [upgrade_type={self.upgrade_type}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .constants import *
from .memo import Memo
from .muxed_account import MuxedAccount
from .operation import Operation
from .preconditions import Preconditions
from .sequence_number import SequenceNumber
from .transaction_ext import TransactionExt
from .uint32 import Uint32

__all__ = ["Transaction"]


class Transaction:
    """
    XDR Source Code::

        struct Transaction
        {
            // account used to run the transaction
            MuxedAccount sourceAccount;

            // the fee the sourceAccount will pay
            uint32 fee;

            // sequence number to consume in the account
            SequenceNumber seqNum;

            // validity conditions
            Preconditions cond;

            Memo memo;

            Operation operations<MAX_OPS_PER_TX>;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            case 1:
                SorobanTransactionData sorobanData;
            }
            ext;
        };
    """

    def __init__(
        self,
        source_account: MuxedAccount,
        fee: Uint32,
        seq_num: SequenceNumber,
        cond: Preconditions,
        memo: Memo,
        operations: List[Operation],
        ext: TransactionExt,
    ) -> None:
        _expect_max_length = MAX_OPS_PER_TX
        if operations and len(operations) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `operations` should be {_expect_max_length}, but got {len(operations)}."
            )
        self.source_account = source_account
        self.fee = fee
        self.seq_num = seq_num
        self.cond = cond
        self.memo = memo
        self.operations = operations
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.source_account.pack(packer)
        self.fee.pack(packer)
        self.seq_num.pack(packer)
        self.cond.pack(packer)
        self.memo.pack(packer)
        packer.pack_uint(len(self.operations))
        for operations_item in self.operations:
            operations_item.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Transaction:
        source_account = MuxedAccount.unpack(unpacker)
        fee = Uint32.unpack(unpacker)
        seq_num = SequenceNumber.unpack(unpacker)
        cond = Preconditions.unpack(unpacker)
        memo = Memo.unpack(unpacker)
        length = unpacker.unpack_uint()
        operations = []
        for _ in range(length):
            operations.append(Operation.unpack(unpacker))
        ext = TransactionExt.unpack(unpacker)
        return cls(
            source_account=source_account,
            fee=fee,
            seq_num=seq_num,
            cond=cond,
            memo=memo,
            operations=operations,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Transaction:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Transaction:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.source_account,
                self.fee,
                self.seq_num,
                self.cond,
                self.memo,
                self.operations,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.source_account == other.source_account
            and self.fee == other.fee
            and self.seq_num == other.seq_num
            and self.cond == other.cond
            and self.memo == other.memo
            and self.operations == other.operations
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"source_account={self.source_account}",
            f"fee={self.fee}",
            f"seq_num={self.seq_num}",
            f"cond={self.cond}",
            f"memo={self.memo}",
            f"operations={self.operations}",
            f"ext={self.ext}",
        ]
        return f"<Transaction [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_entry_type import LedgerEntryType
from .ledger_key_account import LedgerKeyAccount
from .ledger_key_claimable_balance import LedgerKeyClaimableBalance
from .ledger_key_config_setting import LedgerKeyConfigSetting
from .ledger_key_contract_code import LedgerKeyContractCode
from .ledger_key_contract_data import LedgerKeyContractData
from .ledger_key_data import LedgerKeyData
from .ledger_key_liquidity_pool import LedgerKeyLiquidityPool
from .ledger_key_offer import LedgerKeyOffer
from .ledger_key_trust_line import LedgerKeyTrustLine
from .ledger_key_ttl import LedgerKeyTtl

__all__ = ["LedgerKey"]


class LedgerKey:
    """
    XDR Source Code::

        union LedgerKey switch (LedgerEntryType type)
        {
        case ACCOUNT:
            struct
            {
                AccountID accountID;
            } account;

        case TRUSTLINE:
            struct
            {
                AccountID accountID;
                TrustLineAsset asset;
            } trustLine;

        case OFFER:
            struct
            {
                AccountID sellerID;
                int64 offerID;
            } offer;

        case DATA:
            struct
            {
                AccountID accountID;
                string64 dataName;
            } data;

        case CLAIMABLE_BALANCE:
            struct
            {
                ClaimableBalanceID balanceID;
            } claimableBalance;

        case LIQUIDITY_POOL:
            struct
            {
                PoolID liquidityPoolID;
            } liquidityPool;
        case CONTRACT_DATA:
            struct
            {
                SCAddress contract;
                SCVal key;
                ContractDataDurability durability;
            } contractData;
        case CONTRACT_CODE:
            struct
            {
                Hash hash;
            } contractCode;
        case CONFIG_SETTING:
            struct
            {
                ConfigSettingID configSettingID;
            } configSetting;
        case TTL:
            struct
            {
                // Hash of the LedgerKey that is associated with this TTLEntry
                Hash keyHash;
            } ttl;
        };
    """

    def __init__(
        self,
        type: LedgerEntryType,
        account: LedgerKeyAccount = None,
        trust_line: LedgerKeyTrustLine = None,
        offer: LedgerKeyOffer = None,
        data: LedgerKeyData = None,
        claimable_balance: LedgerKeyClaimableBalance = None,
        liquidity_pool: LedgerKeyLiquidityPool = None,
        contract_data: LedgerKeyContractData = None,
        contract_code: LedgerKeyContractCode = None,
        config_setting: LedgerKeyConfigSetting = None,
        ttl: LedgerKeyTtl = None,
    ) -> None:
        self.type = type
        self.account = account
        self.trust_line = trust_line
        self.offer = offer
        self.data = data
        self.claimable_balance = claimable_balance
        self.liquidity_pool = liquidity_pool
        self.contract_data = contract_data
        self.contract_code = contract_code
        self.config_setting = config_setting
        self.ttl = ttl

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == LedgerEntryType.ACCOUNT:
            if self.account is None:
                raise ValueError("account should not be None.")
            self.account.pack(packer)
            return
        if self.type == LedgerEntryType.TRUSTLINE:
            if self.trust_line is None:
                raise ValueError("trust_line should not be None.")
            self.trust_line.pack(packer)
            return
        if self.type == LedgerEntryType.OFFER:
            if self.offer is None:
                raise ValueError("offer should not be None.")
            self.offer.pack(packer)
            return
        if self.type == LedgerEntryType.DATA:
            if self.data is None:
                raise ValueError("data should not be None.")
            self.data.pack(packer)
            return
        if self.type == LedgerEntryType.CLAIMABLE_BALANCE:
            if self.claimable_balance is None:
                raise ValueError("claimable_balance should not be None.")
            self.claimable_balance.pack(packer)
            return
        if self.type == LedgerEntryType.LIQUIDITY_POOL:
            if self.liquidity_pool is None:
                raise ValueError("liquidity_pool should not be None.")
            self.liquidity_pool.pack(packer)
            return
        if self.type == LedgerEntryType.CONTRACT_DATA:
            if self.contract_data is None:
                raise ValueError("contract_data should not be None.")
            self.contract_data.pack(packer)
            return
        if self.type == LedgerEntryType.CONTRACT_CODE:
            if self.contract_code is None:
                raise ValueError("contract_code should not be None.")
            self.contract_code.pack(packer)
            return
        if self.type == LedgerEntryType.CONFIG_SETTING:
            if self.config_setting is None:
                raise ValueError("config_setting should not be None.")
            self.config_setting.pack(packer)
            return
        if self.type == LedgerEntryType.TTL:
            if self.ttl is None:
                raise ValueError("ttl should not be None.")
            self.ttl.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerKey:
        type = LedgerEntryType.unpack(unpacker)
        if type == LedgerEntryType.ACCOUNT:
            account = LedgerKeyAccount.unpack(unpacker)
            return cls(type=type, account=account)
        if type == LedgerEntryType.TRUSTLINE:
            trust_line = LedgerKeyTrustLine.unpack(unpacker)
            return cls(type=type, trust_line=trust_line)
        if type == LedgerEntryType.OFFER:
            offer = LedgerKeyOffer.unpack(unpacker)
            return cls(type=type, offer=offer)
        if type == LedgerEntryType.DATA:
            data = LedgerKeyData.unpack(unpacker)
            return cls(type=type, data=data)
        if type == LedgerEntryType.CLAIMABLE_BALANCE:
            claimable_balance = LedgerKeyClaimableBalance.unpack(unpacker)
            return cls(type=type, claimable_balance=claimable_balance)
        if type == LedgerEntryType.LIQUIDITY_POOL:
            liquidity_pool = LedgerKeyLiquidityPool.unpack(unpacker)
            return cls(type=type, liquidity_pool=liquidity_pool)
        if type == LedgerEntryType.CONTRACT_DATA:
            contract_data = LedgerKeyContractData.unpack(unpacker)
            return cls(type=type, contract_data=contract_data)
        if type == LedgerEntryType.CONTRACT_CODE:
            contract_code = LedgerKeyContractCode.unpack(unpacker)
            return cls(type=type, contract_code=contract_code)
        if type == LedgerEntryType.CONFIG_SETTING:
            config_setting = LedgerKeyConfigSetting.unpack(unpacker)
            return cls(type=type, config_setting=config_setting)
        if type == LedgerEntryType.TTL:
            ttl = LedgerKeyTtl.unpack(unpacker)
            return cls(type=type, ttl=ttl)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerKey:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerKey:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.account,
                self.trust_line,
                self.offer,
                self.data,
                self.claimable_balance,
                self.liquidity_pool,
                self.contract_data,
                self.contract_code,
                self.config_setting,
                self.ttl,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.account == other.account
            and self.trust_line == other.trust_line
            and self.offer == other.offer
            and self.data == other.data
            and self.claimable_balance == other.claimable_balance
            and self.liquidity_pool == other.liquidity_pool
            and self.contract_data == other.contract_data
            and self.contract_code == other.contract_code
            and self.config_setting == other.config_setting
            and self.ttl == other.ttl
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"account={self.account}") if self.account is not None else None
        out.append(
            f"trust_line={self.trust_line}"
        ) if self.trust_line is not None else None
        out.append(f"offer={self.offer}") if self.offer is not None else None
        out.append(f"data={self.data}") if self.data is not None else None
        out.append(
            f"claimable_balance={self.claimable_balance}"
        ) if self.claimable_balance is not None else None
        out.append(
            f"liquidity_pool={self.liquidity_pool}"
        ) if self.liquidity_pool is not None else None
        out.append(
            f"contract_data={self.contract_data}"
        ) if self.contract_data is not None else None
        out.append(
            f"contract_code={self.contract_code}"
        ) if self.contract_code is not None else None
        out.append(
            f"config_setting={self.config_setting}"
        ) if self.config_setting is not None else None
        out.append(f"ttl={self.ttl}") if self.ttl is not None else None
        return f"<LedgerKey [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .signature import Signature
from .survey_request_message import SurveyRequestMessage

__all__ = ["SignedSurveyRequestMessage"]


class SignedSurveyRequestMessage:
    """
    XDR Source Code::

        struct SignedSurveyRequestMessage
        {
            Signature requestSignature;
            SurveyRequestMessage request;
        };
    """

    def __init__(
        self,
        request_signature: Signature,
        request: SurveyRequestMessage,
    ) -> None:
        self.request_signature = request_signature
        self.request = request

    def pack(self, packer: Packer) -> None:
        self.request_signature.pack(packer)
        self.request.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SignedSurveyRequestMessage:
        request_signature = Signature.unpack(unpacker)
        request = SurveyRequestMessage.unpack(unpacker)
        return cls(
            request_signature=request_signature,
            request=request,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SignedSurveyRequestMessage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SignedSurveyRequestMessage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.request_signature,
                self.request,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.request_signature == other.request_signature
            and self.request == other.request
        )

    def __str__(self):
        out = [
            f"request_signature={self.request_signature}",
            f"request={self.request}",
        ]
        return f"<SignedSurveyRequestMessage [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .uint32 import Uint32

__all__ = ["LedgerBounds"]


class LedgerBounds:
    """
    XDR Source Code::

        struct LedgerBounds
        {
            uint32 minLedger;
            uint32 maxLedger; // 0 here means no maxLedger
        };
    """

    def __init__(
        self,
        min_ledger: Uint32,
        max_ledger: Uint32,
    ) -> None:
        self.min_ledger = min_ledger
        self.max_ledger = max_ledger

    def pack(self, packer: Packer) -> None:
        self.min_ledger.pack(packer)
        self.max_ledger.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerBounds:
        min_ledger = Uint32.unpack(unpacker)
        max_ledger = Uint32.unpack(unpacker)
        return cls(
            min_ledger=min_ledger,
            max_ledger=max_ledger,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerBounds:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerBounds:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.min_ledger,
                self.max_ledger,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.min_ledger == other.min_ledger and self.max_ledger == other.max_ledger
        )

    def __str__(self):
        out = [
            f"min_ledger={self.min_ledger}",
            f"max_ledger={self.max_ledger}",
        ]
        return f"<LedgerBounds [{', '.join(out)}]>"



# Automatically generated by xdrgen
# DO NOT EDIT or your changes may be overwritten
from .account_entry import AccountEntry
from .account_entry_ext import AccountEntryExt
from .account_entry_extension_v1 import AccountEntryExtensionV1
from .account_entry_extension_v1_ext import AccountEntryExtensionV1Ext
from .account_entry_extension_v2 import AccountEntryExtensionV2
from .account_entry_extension_v2_ext import AccountEntryExtensionV2Ext
from .account_entry_extension_v3 import AccountEntryExtensionV3
from .account_flags import AccountFlags
from .account_id import AccountID
from .account_merge_result import AccountMergeResult
from .account_merge_result_code import AccountMergeResultCode
from .allow_trust_op import AllowTrustOp
from .allow_trust_result import AllowTrustResult
from .allow_trust_result_code import AllowTrustResultCode
from .alpha_num4 import AlphaNum4
from .alpha_num12 import AlphaNum12
from .asset import Asset
from .asset_code import AssetCode
from .asset_code4 import AssetCode4
from .asset_code12 import AssetCode12
from .asset_type import AssetType
from .auth import Auth
from .auth_cert import AuthCert
from .authenticated_message import AuthenticatedMessage
from .authenticated_message_v0 import AuthenticatedMessageV0
from .base import *
from .begin_sponsoring_future_reserves_op import BeginSponsoringFutureReservesOp
from .begin_sponsoring_future_reserves_result import BeginSponsoringFutureReservesResult
from .begin_sponsoring_future_reserves_result_code import (
    BeginSponsoringFutureReservesResultCode,
)
from .bucket_entry import BucketEntry
from .bucket_entry_type import BucketEntryType
from .bucket_metadata import BucketMetadata
from .bucket_metadata_ext import BucketMetadataExt
from .bump_sequence_op import BumpSequenceOp
from .bump_sequence_result import BumpSequenceResult
from .bump_sequence_result_code import BumpSequenceResultCode
from .change_trust_asset import ChangeTrustAsset
from .change_trust_op import ChangeTrustOp
from .change_trust_result import ChangeTrustResult
from .change_trust_result_code import ChangeTrustResultCode
from .claim_atom import ClaimAtom
from .claim_atom_type import ClaimAtomType
from .claim_claimable_balance_op import ClaimClaimableBalanceOp
from .claim_claimable_balance_result import ClaimClaimableBalanceResult
from .claim_claimable_balance_result_code import ClaimClaimableBalanceResultCode
from .claim_liquidity_atom import ClaimLiquidityAtom
from .claim_offer_atom import ClaimOfferAtom
from .claim_offer_atom_v0 import ClaimOfferAtomV0
from .claim_predicate import ClaimPredicate
from .claim_predicate_type import ClaimPredicateType
from .claimable_balance_entry import ClaimableBalanceEntry
from .claimable_balance_entry_ext import ClaimableBalanceEntryExt
from .claimable_balance_entry_extension_v1 import ClaimableBalanceEntryExtensionV1
from .claimable_balance_entry_extension_v1_ext import (
    ClaimableBalanceEntryExtensionV1Ext,
)
from .claimable_balance_flags import ClaimableBalanceFlags
from .claimable_balance_id import ClaimableBalanceID
from .claimable_balance_id_type import ClaimableBalanceIDType
from .claimant import Claimant
from .claimant_type import ClaimantType
from .claimant_v0 import ClaimantV0
from .clawback_claimable_balance_op import ClawbackClaimableBalanceOp
from .clawback_claimable_balance_result import ClawbackClaimableBalanceResult
from .clawback_claimable_balance_result_code import ClawbackClaimableBalanceResultCode
from .clawback_op import ClawbackOp
from .clawback_result import ClawbackResult
from .clawback_result_code import ClawbackResultCode
from .config_setting_contract_bandwidth_v0 import ConfigSettingContractBandwidthV0
from .config_setting_contract_compute_v0 import ConfigSettingContractComputeV0
from .config_setting_contract_events_v0 import ConfigSettingContractEventsV0
from .config_setting_contract_execution_lanes_v0 import (
    ConfigSettingContractExecutionLanesV0,
)
from .config_setting_contract_historical_data_v0 import (
    ConfigSettingContractHistoricalDataV0,
)
from .config_setting_contract_ledger_cost_v0 import ConfigSettingContractLedgerCostV0
from .config_setting_entry import ConfigSettingEntry
from .config_setting_id import ConfigSettingID
from .config_upgrade_set import ConfigUpgradeSet
from .config_upgrade_set_key import ConfigUpgradeSetKey
from .constants import *
from .contract_code_entry import ContractCodeEntry
from .contract_cost_param_entry import ContractCostParamEntry
from .contract_cost_params import ContractCostParams
from .contract_cost_type import ContractCostType
from .contract_data_durability import ContractDataDurability
from .contract_data_entry import ContractDataEntry
from .contract_event import ContractEvent
from .contract_event_body import ContractEventBody
from .contract_event_type import ContractEventType
from .contract_event_v0 import ContractEventV0
from .contract_executable import ContractExecutable
from .contract_executable_type import ContractExecutableType
from .contract_id_preimage import ContractIDPreimage
from .contract_id_preimage_from_address import ContractIDPreimageFromAddress
from .contract_id_preimage_type import ContractIDPreimageType
from .create_account_op import CreateAccountOp
from .create_account_result import CreateAccountResult
from .create_account_result_code import CreateAccountResultCode
from .create_claimable_balance_op import CreateClaimableBalanceOp
from .create_claimable_balance_result import CreateClaimableBalanceResult
from .create_claimable_balance_result_code import CreateClaimableBalanceResultCode
from .create_contract_args import CreateContractArgs
from .create_passive_sell_offer_op import CreatePassiveSellOfferOp
from .crypto_key_type import CryptoKeyType
from .curve25519_public import Curve25519Public
from .curve25519_secret import Curve25519Secret
from .data_entry import DataEntry
from .data_entry_ext import DataEntryExt
from .data_value import DataValue
from .decorated_signature import DecoratedSignature
from .diagnostic_event import DiagnosticEvent
from .dont_have import DontHave
from .duration import Duration
from .encrypted_body import EncryptedBody
from .end_sponsoring_future_reserves_result import EndSponsoringFutureReservesResult
from .end_sponsoring_future_reserves_result_code import (
    EndSponsoringFutureReservesResultCode,
)
from .envelope_type import EnvelopeType
from .error import Error
from .error_code import ErrorCode
from .eviction_iterator import EvictionIterator
from .extend_footprint_ttl_op import ExtendFootprintTTLOp
from .extend_footprint_ttl_result import ExtendFootprintTTLResult
from .extend_footprint_ttl_result_code import ExtendFootprintTTLResultCode
from .extension_point import ExtensionPoint
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .fee_bump_transaction_ext import FeeBumpTransactionExt
from .fee_bump_transaction_inner_tx import FeeBumpTransactionInnerTx
from .flood_advert import FloodAdvert
from .flood_demand import FloodDemand
from .generalized_transaction_set import GeneralizedTransactionSet
from .hash import Hash
from .hash_id_preimage import HashIDPreimage
from .hash_id_preimage_contract_id import HashIDPreimageContractID
from .hash_id_preimage_operation_id import HashIDPreimageOperationID
from .hash_id_preimage_revoke_id import HashIDPreimageRevokeID
from .hash_id_preimage_soroban_authorization import HashIDPreimageSorobanAuthorization
from .hello import Hello
from .hmac_sha256_key import HmacSha256Key
from .hmac_sha256_mac import HmacSha256Mac
from .host_function import HostFunction
from .host_function_type import HostFunctionType
from .inflation_payout import InflationPayout
from .inflation_result import InflationResult
from .inflation_result_code import InflationResultCode
from .inner_transaction_result import InnerTransactionResult
from .inner_transaction_result_ext import InnerTransactionResultExt
from .inner_transaction_result_pair import InnerTransactionResultPair
from .inner_transaction_result_result import InnerTransactionResultResult
from .int32 import Int32
from .int64 import Int64
from .int128_parts import Int128Parts
from .int256_parts import Int256Parts
from .invoke_contract_args import InvokeContractArgs
from .invoke_host_function_op import InvokeHostFunctionOp
from .invoke_host_function_result import InvokeHostFunctionResult
from .invoke_host_function_result_code import InvokeHostFunctionResultCode
from .invoke_host_function_success_pre_image import InvokeHostFunctionSuccessPreImage
from .ip_addr_type import IPAddrType
from .ledger_bounds import LedgerBounds
from .ledger_close_meta import LedgerCloseMeta
from .ledger_close_meta_v0 import LedgerCloseMetaV0
from .ledger_close_meta_v1 import LedgerCloseMetaV1
from .ledger_close_value_signature import LedgerCloseValueSignature
from .ledger_entry import LedgerEntry
from .ledger_entry_change import LedgerEntryChange
from .ledger_entry_change_type import LedgerEntryChangeType
from .ledger_entry_changes import LedgerEntryChanges
from .ledger_entry_data import LedgerEntryData
from .ledger_entry_ext import LedgerEntryExt
from .ledger_entry_extension_v1 import LedgerEntryExtensionV1
from .ledger_entry_extension_v1_ext import LedgerEntryExtensionV1Ext
from .ledger_entry_type import LedgerEntryType
from .ledger_footprint import LedgerFootprint
from .ledger_header import LedgerHeader
from .ledger_header_ext import LedgerHeaderExt
from .ledger_header_extension_v1 import LedgerHeaderExtensionV1
from .ledger_header_extension_v1_ext import LedgerHeaderExtensionV1Ext
from .ledger_header_flags import LedgerHeaderFlags
from .ledger_header_history_entry import LedgerHeaderHistoryEntry
from .ledger_header_history_entry_ext import LedgerHeaderHistoryEntryExt
from .ledger_key import LedgerKey
from .ledger_key_account import LedgerKeyAccount
from .ledger_key_claimable_balance import LedgerKeyClaimableBalance
from .ledger_key_config_setting import LedgerKeyConfigSetting
from .ledger_key_contract_code import LedgerKeyContractCode
from .ledger_key_contract_data import LedgerKeyContractData
from .ledger_key_data import LedgerKeyData
from .ledger_key_liquidity_pool import LedgerKeyLiquidityPool
from .ledger_key_offer import LedgerKeyOffer
from .ledger_key_trust_line import LedgerKeyTrustLine
from .ledger_key_ttl import LedgerKeyTtl
from .ledger_scp_messages import LedgerSCPMessages
from .ledger_upgrade import LedgerUpgrade
from .ledger_upgrade_type import LedgerUpgradeType
from .liabilities import Liabilities
from .liquidity_pool_constant_product_parameters import (
    LiquidityPoolConstantProductParameters,
)
from .liquidity_pool_deposit_op import LiquidityPoolDepositOp
from .liquidity_pool_deposit_result import LiquidityPoolDepositResult
from .liquidity_pool_deposit_result_code import LiquidityPoolDepositResultCode
from .liquidity_pool_entry import LiquidityPoolEntry
from .liquidity_pool_entry_body import LiquidityPoolEntryBody
from .liquidity_pool_entry_constant_product import LiquidityPoolEntryConstantProduct
from .liquidity_pool_parameters import LiquidityPoolParameters
from .liquidity_pool_type import LiquidityPoolType
from .liquidity_pool_withdraw_op import LiquidityPoolWithdrawOp
from .liquidity_pool_withdraw_result import LiquidityPoolWithdrawResult
from .liquidity_pool_withdraw_result_code import LiquidityPoolWithdrawResultCode
from .manage_buy_offer_op import ManageBuyOfferOp
from .manage_buy_offer_result import ManageBuyOfferResult
from .manage_buy_offer_result_code import ManageBuyOfferResultCode
from .manage_data_op import ManageDataOp
from .manage_data_result import ManageDataResult
from .manage_data_result_code import ManageDataResultCode
from .manage_offer_effect import ManageOfferEffect
from .manage_offer_success_result import ManageOfferSuccessResult
from .manage_offer_success_result_offer import ManageOfferSuccessResultOffer
from .manage_sell_offer_op import ManageSellOfferOp
from .manage_sell_offer_result import ManageSellOfferResult
from .manage_sell_offer_result_code import ManageSellOfferResultCode
from .memo import Memo
from .memo_type import MemoType
from .message_type import MessageType
from .muxed_account import MuxedAccount
from .muxed_account_med25519 import MuxedAccountMed25519
from .node_id import NodeID
from .offer_entry import OfferEntry
from .offer_entry_ext import OfferEntryExt
from .offer_entry_flags import OfferEntryFlags
from .operation import Operation
from .operation_body import OperationBody
from .operation_meta import OperationMeta
from .operation_result import OperationResult
from .operation_result_code import OperationResultCode
from .operation_result_tr import OperationResultTr
from .operation_type import OperationType
from .path_payment_strict_receive_op import PathPaymentStrictReceiveOp
from .path_payment_strict_receive_result import PathPaymentStrictReceiveResult
from .path_payment_strict_receive_result_code import PathPaymentStrictReceiveResultCode
from .path_payment_strict_receive_result_success import (
    PathPaymentStrictReceiveResultSuccess,
)
from .path_payment_strict_send_op import PathPaymentStrictSendOp
from .path_payment_strict_send_result import PathPaymentStrictSendResult
from .path_payment_strict_send_result_code import PathPaymentStrictSendResultCode
from .path_payment_strict_send_result_success import PathPaymentStrictSendResultSuccess
from .payment_op import PaymentOp
from .payment_result import PaymentResult
from .payment_result_code import PaymentResultCode
from .peer_address import PeerAddress
from .peer_address_ip import PeerAddressIp
from .peer_stat_list import PeerStatList
from .peer_stats import PeerStats
from .persisted_scp_state import PersistedSCPState
from .persisted_scp_state_v0 import PersistedSCPStateV0
from .persisted_scp_state_v1 import PersistedSCPStateV1
from .pool_id import PoolID
from .precondition_type import PreconditionType
from .preconditions import Preconditions
from .preconditions_v2 import PreconditionsV2
from .price import Price
from .public_key import PublicKey
from .public_key_type import PublicKeyType
from .restore_footprint_op import RestoreFootprintOp
from .restore_footprint_result import RestoreFootprintResult
from .restore_footprint_result_code import RestoreFootprintResultCode
from .revoke_sponsorship_op import RevokeSponsorshipOp
from .revoke_sponsorship_op_signer import RevokeSponsorshipOpSigner
from .revoke_sponsorship_result import RevokeSponsorshipResult
from .revoke_sponsorship_result_code import RevokeSponsorshipResultCode
from .revoke_sponsorship_type import RevokeSponsorshipType
from .sc_address import SCAddress
from .sc_address_type import SCAddressType
from .sc_bytes import SCBytes
from .sc_contract_instance import SCContractInstance
from .sc_env_meta_entry import SCEnvMetaEntry
from .sc_env_meta_kind import SCEnvMetaKind
from .sc_error import SCError
from .sc_error_code import SCErrorCode
from .sc_error_type import SCErrorType
from .sc_map import SCMap
from .sc_map_entry import SCMapEntry
from .sc_meta_entry import SCMetaEntry
from .sc_meta_kind import SCMetaKind
from .sc_meta_v0 import SCMetaV0
from .sc_nonce_key import SCNonceKey
from .sc_spec_entry import SCSpecEntry
from .sc_spec_entry_kind import SCSpecEntryKind
from .sc_spec_function_input_v0 import SCSpecFunctionInputV0
from .sc_spec_function_v0 import SCSpecFunctionV0
from .sc_spec_type import SCSpecType
from .sc_spec_type_bytes_n import SCSpecTypeBytesN
from .sc_spec_type_def import SCSpecTypeDef
from .sc_spec_type_map import SCSpecTypeMap
from .sc_spec_type_option import SCSpecTypeOption
from .sc_spec_type_result import SCSpecTypeResult
from .sc_spec_type_tuple import SCSpecTypeTuple
from .sc_spec_type_udt import SCSpecTypeUDT
from .sc_spec_type_vec import SCSpecTypeVec
from .sc_spec_udt_enum_case_v0 import SCSpecUDTEnumCaseV0
from .sc_spec_udt_enum_v0 import SCSpecUDTEnumV0
from .sc_spec_udt_error_enum_case_v0 import SCSpecUDTErrorEnumCaseV0
from .sc_spec_udt_error_enum_v0 import SCSpecUDTErrorEnumV0
from .sc_spec_udt_struct_field_v0 import SCSpecUDTStructFieldV0
from .sc_spec_udt_struct_v0 import SCSpecUDTStructV0
from .sc_spec_udt_union_case_tuple_v0 import SCSpecUDTUnionCaseTupleV0
from .sc_spec_udt_union_case_v0 import SCSpecUDTUnionCaseV0
from .sc_spec_udt_union_case_v0_kind import SCSpecUDTUnionCaseV0Kind
from .sc_spec_udt_union_case_void_v0 import SCSpecUDTUnionCaseVoidV0
from .sc_spec_udt_union_v0 import SCSpecUDTUnionV0
from .sc_string import SCString
from .sc_symbol import SCSymbol
from .sc_val import SCVal
from .sc_val_type import SCValType
from .sc_vec import SCVec
from .scp_ballot import SCPBallot
from .scp_envelope import SCPEnvelope
from .scp_history_entry import SCPHistoryEntry
from .scp_history_entry_v0 import SCPHistoryEntryV0
from .scp_nomination import SCPNomination
from .scp_quorum_set import SCPQuorumSet
from .scp_statement import SCPStatement
from .scp_statement_confirm import SCPStatementConfirm
from .scp_statement_externalize import SCPStatementExternalize
from .scp_statement_pledges import SCPStatementPledges
from .scp_statement_prepare import SCPStatementPrepare
from .scp_statement_type import SCPStatementType
from .send_more import SendMore
from .send_more_extended import SendMoreExtended
from .sequence_number import SequenceNumber
from .set_options_op import SetOptionsOp
from .set_options_result import SetOptionsResult
from .set_options_result_code import SetOptionsResultCode
from .set_trust_line_flags_op import SetTrustLineFlagsOp
from .set_trust_line_flags_result import SetTrustLineFlagsResult
from .set_trust_line_flags_result_code import SetTrustLineFlagsResultCode
from .signature import Signature
from .signature_hint import SignatureHint
from .signed_survey_request_message import SignedSurveyRequestMessage
from .signed_survey_response_message import SignedSurveyResponseMessage
from .signer import Signer
from .signer_key import SignerKey
from .signer_key_ed25519_signed_payload import SignerKeyEd25519SignedPayload
from .signer_key_type import SignerKeyType
from .simple_payment_result import SimplePaymentResult
from .soroban_address_credentials import SorobanAddressCredentials
from .soroban_authorization_entry import SorobanAuthorizationEntry
from .soroban_authorized_function import SorobanAuthorizedFunction
from .soroban_authorized_function_type import SorobanAuthorizedFunctionType
from .soroban_authorized_invocation import SorobanAuthorizedInvocation
from .soroban_credentials import SorobanCredentials
from .soroban_credentials_type import SorobanCredentialsType
from .soroban_resources import SorobanResources
from .soroban_transaction_data import SorobanTransactionData
from .soroban_transaction_meta import SorobanTransactionMeta
from .sponsorship_descriptor import SponsorshipDescriptor
from .state_archival_settings import StateArchivalSettings
from .stellar_message import StellarMessage
from .stellar_value import StellarValue
from .stellar_value_ext import StellarValueExt
from .stellar_value_type import StellarValueType
from .stored_debug_transaction_set import StoredDebugTransactionSet
from .stored_transaction_set import StoredTransactionSet
from .string32 import String32
from .string64 import String64
from .survey_message_command_type import SurveyMessageCommandType
from .survey_message_response_type import SurveyMessageResponseType
from .survey_request_message import SurveyRequestMessage
from .survey_response_body import SurveyResponseBody
from .survey_response_message import SurveyResponseMessage
from .threshold_indexes import ThresholdIndexes
from .thresholds import Thresholds
from .time_bounds import TimeBounds
from .time_point import TimePoint
from .topology_response_body_v0 import TopologyResponseBodyV0
from .topology_response_body_v1 import TopologyResponseBodyV1
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope
from .transaction_ext import TransactionExt
from .transaction_history_entry import TransactionHistoryEntry
from .transaction_history_entry_ext import TransactionHistoryEntryExt
from .transaction_history_result_entry import TransactionHistoryResultEntry
from .transaction_history_result_entry_ext import TransactionHistoryResultEntryExt
from .transaction_meta import TransactionMeta
from .transaction_meta_v1 import TransactionMetaV1
from .transaction_meta_v2 import TransactionMetaV2
from .transaction_meta_v3 import TransactionMetaV3
from .transaction_phase import TransactionPhase
from .transaction_result import TransactionResult
from .transaction_result_code import TransactionResultCode
from .transaction_result_ext import TransactionResultExt
from .transaction_result_meta import TransactionResultMeta
from .transaction_result_pair import TransactionResultPair
from .transaction_result_result import TransactionResultResult
from .transaction_result_set import TransactionResultSet
from .transaction_set import TransactionSet
from .transaction_set_v1 import TransactionSetV1
from .transaction_signature_payload import TransactionSignaturePayload
from .transaction_signature_payload_tagged_transaction import (
    TransactionSignaturePayloadTaggedTransaction,
)
from .transaction_v0 import TransactionV0
from .transaction_v0_envelope import TransactionV0Envelope
from .transaction_v0_ext import TransactionV0Ext
from .transaction_v1_envelope import TransactionV1Envelope
from .trust_line_asset import TrustLineAsset
from .trust_line_entry import TrustLineEntry
from .trust_line_entry_ext import TrustLineEntryExt
from .trust_line_entry_extension_v2 import TrustLineEntryExtensionV2
from .trust_line_entry_extension_v2_ext import TrustLineEntryExtensionV2Ext
from .trust_line_entry_v1 import TrustLineEntryV1
from .trust_line_entry_v1_ext import TrustLineEntryV1Ext
from .trust_line_flags import TrustLineFlags
from .ttl_entry import TTLEntry
from .tx_advert_vector import TxAdvertVector
from .tx_demand_vector import TxDemandVector
from .tx_set_component import TxSetComponent
from .tx_set_component_txs_maybe_discounted_fee import (
    TxSetComponentTxsMaybeDiscountedFee,
)
from .tx_set_component_type import TxSetComponentType
from .u_int128_parts import UInt128Parts
from .u_int256_parts import UInt256Parts
from .uint32 import Uint32
from .uint64 import Uint64
from .uint256 import Uint256
from .upgrade_entry_meta import UpgradeEntryMeta
from .upgrade_type import UpgradeType
from .value import Value



import asyncio
import json
import logging
from typing import Any, AsyncGenerator, Dict, Optional

from ..__version__ import __version__
from ..exceptions import ConnectionError, StreamClientError
from . import defines
from .base_async_client import BaseAsyncClient
from .response import Response

logger = logging.getLogger(__name__)

DEFAULT_NUM_RETRIES = 3
DEFAULT_BACKOFF_FACTOR = 0.5
USER_AGENT = f"py-stellar-base/{__version__}/AiohttpClient"
IDENTIFICATION_HEADERS = {
    "X-Client-Name": "py-stellar-base",
    "X-Client-Version": __version__,
}

__all__ = ["AiohttpClient"]


# Temporary solution to solve the problem that
# `aiohttp_sse_client` cannot read long stream messages.
# Let's rewrite a stream lib later.
async def __readline(self) -> bytes:
    if self._exception is not None:
        raise self._exception

    line = []
    line_size = 0
    not_enough = True

    while not_enough:
        while self._buffer and not_enough:
            offset = self._buffer_offset
            ichar = self._buffer[0].find(b"\n", offset) + 1
            # Read from current offset to found b'\n' or to the end.
            data = self._read_nowait_chunk(ichar - offset if ichar else -1)
            line.append(data)
            line_size += len(data)
            if ichar:
                not_enough = False
            #
            # if line_size > self._high_water:
            #     raise ValueError('Line is too long')
        if self._eof:
            break

        if not_enough:
            await self._wait("readline")

    return b"".join(line)


try:
    import aiohttp
    from aiohttp_sse_client.client import EventSource

    aiohttp.streams.StreamReader.readline = __readline  # type: ignore[assignment]
    _AIOHTTP_DEPS_INSTALLED = True
except ImportError:
    _AIOHTTP_DEPS_INSTALLED = False


class AiohttpClient(BaseAsyncClient):
    """The :class:`AiohttpClient` object is a asynchronous http client,
    which represents the interface for making requests to a server instance.

    :param pool_size: persistent connection to Horizon and connection pool
    :param request_timeout: the timeout for all GET requests
    :param post_timeout: the timeout for all POST requests
    :param backoff_factor: a backoff factor to apply between attempts after the second try
    :param user_agent: the server can use it to identify you
    :param custom_headers: any additional HTTP headers to add in requests
    """

    def __init__(
        self,
        pool_size: Optional[int] = None,
        request_timeout: float = defines.DEFAULT_GET_TIMEOUT_SECONDS,
        post_timeout: float = defines.DEFAULT_POST_TIMEOUT_SECONDS,
        backoff_factor: Optional[float] = DEFAULT_BACKOFF_FACTOR,
        user_agent: Optional[str] = None,
        custom_headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> None:
        if not _AIOHTTP_DEPS_INSTALLED:
            raise ImportError(
                "The required dependencies have not been installed. "
                "Please install `stellar-sdk[aiohttp]` to use this feature."
            )
        self.pool_size = pool_size
        self.backoff_factor: Optional[float] = backoff_factor
        self.request_timeout: float = request_timeout
        self.post_timeout: float = post_timeout
        # init session
        if pool_size is None:
            connector = aiohttp.TCPConnector()
        else:
            connector = aiohttp.TCPConnector(limit=pool_size)

        self.user_agent: Optional[str] = USER_AGENT
        if user_agent:
            self.user_agent = user_agent

        self.headers: dict = {
            **IDENTIFICATION_HEADERS,
            "User-Agent": self.user_agent,
        }

        if custom_headers:
            self.headers = {**self.headers, **custom_headers}

        session = aiohttp.ClientSession(
            headers=self.headers.copy(),
            connector=connector,
            timeout=aiohttp.ClientTimeout(total=request_timeout),
            **kwargs,
        )

        self._session: aiohttp.ClientSession = session
        self._sse_session: Optional[aiohttp.ClientSession] = None

    async def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            response = await self._session.get(url, params=params)
            return Response(
                status_code=response.status,
                text=await response.text(),
                headers=dict(response.headers),
                url=str(response.url),
            )
        except aiohttp.ClientError as e:  # TODO: need more research
            raise ConnectionError(e)

    async def post(
        self, url: str, data: Dict[str, str] = None, json_data: Dict[str, Any] = None
    ) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :param json_data: the json data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            response = await self._session.post(
                url,
                data=data,
                json=json_data,
                timeout=aiohttp.ClientTimeout(total=self.post_timeout),
            )
            return Response(
                status_code=response.status,
                text=await response.text(),
                headers=dict(response.headers),
                url=str(response.url),
            )
        except aiohttp.ClientError as e:
            raise ConnectionError(e)

    async def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """Perform Stream request.

        :param url: the request url
        :param params: the request params
        :return: the stream response from server
        :raise: :exc:`StreamClientError <stellar_sdk.exceptions.StreamClientError>` - Failed to fetch stream resource.
        """

        # Init the sse session
        if self._sse_session is None:
            # No special connector
            # Other headers such as "Accept: text/event-stream" are added by thr SSEClient
            timeout = aiohttp.ClientTimeout(total=60 * 5)
            self._sse_session = aiohttp.ClientSession(timeout=timeout)

        query_params = {**params} if params else dict()

        query_params.update(**IDENTIFICATION_HEADERS)
        retry = 0.1

        while True:
            try:
                """
                Create a new SSEClient:
                Using the last id as the cursor
                Headers are needed because of a bug that makes "params" override the default headers
                """
                async with EventSource(
                    url,
                    session=self._sse_session,
                    params=query_params,
                    headers=self.headers.copy(),
                ) as client:
                    """
                    We want to throw a TimeoutError if we didnt get any event in the last x seconds.
                    read_timeout in aiohttp is not implemented correctly https://github.com/aio-libs/aiohttp/issues/1954
                    So we will create our own way to do that.

                    Note that the timeout starts from the first event forward. There is no until we get the first event.
                    """
                    async for event in client:
                        if event.last_event_id:
                            query_params["cursor"] = event.last_event_id
                            # Events that dont have an id are not useful for us (hello/byebye events)
                        retry = client._reconnection_time.total_seconds()
                        try:
                            data = event.data
                            if data != '"hello"' and data != '"byebye"':
                                yield json.loads(data)
                        except json.JSONDecodeError:
                            # Content was not json-decodable
                            pass  # pragma: no cover
            except aiohttp.ClientError as e:
                raise StreamClientError(
                    query_params["cursor"], "Failed to get stream message."
                ) from e
            except asyncio.TimeoutError:
                logger.warning(
                    f"We have encountered an timeout error and we will try to reconnect, cursor = {query_params.get('cursor')}"
                )
                await asyncio.sleep(retry)

    async def __aenter__(self) -> "AiohttpClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def close(self) -> None:
        """Close underlying connector.

        Release all acquired resources.
        """
        await self._session.__aexit__(None, None, None)
        if self._sse_session is not None:
            await self._sse_session.__aexit__(None, None, None)

    def __str__(self):
        return (
            f"<AiohttpClient [pool_size={self.pool_size}, "
            f"request_timeout={self.request_timeout}, "
            f"post_timeout={self.post_timeout}, "
            f"backoff_factor={self.backoff_factor}, "
            f"user_agent={self.user_agent}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_env_meta_kind import SCEnvMetaKind
from .uint64 import Uint64

__all__ = ["SCEnvMetaEntry"]


class SCEnvMetaEntry:
    """
    XDR Source Code::

        union SCEnvMetaEntry switch (SCEnvMetaKind kind)
        {
        case SC_ENV_META_KIND_INTERFACE_VERSION:
            uint64 interfaceVersion;
        };
    """

    def __init__(
        self,
        kind: SCEnvMetaKind,
        interface_version: Uint64 = None,
    ) -> None:
        self.kind = kind
        self.interface_version = interface_version

    def pack(self, packer: Packer) -> None:
        self.kind.pack(packer)
        if self.kind == SCEnvMetaKind.SC_ENV_META_KIND_INTERFACE_VERSION:
            if self.interface_version is None:
                raise ValueError("interface_version should not be None.")
            self.interface_version.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCEnvMetaEntry:
        kind = SCEnvMetaKind.unpack(unpacker)
        if kind == SCEnvMetaKind.SC_ENV_META_KIND_INTERFACE_VERSION:
            interface_version = Uint64.unpack(unpacker)
            return cls(kind=kind, interface_version=interface_version)
        return cls(kind=kind)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCEnvMetaEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCEnvMetaEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.kind,
                self.interface_version,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.kind == other.kind
            and self.interface_version == other.interface_version
        )

    def __str__(self):
        out = []
        out.append(f"kind={self.kind}")
        out.append(
            f"interface_version={self.interface_version}"
        ) if self.interface_version is not None else None
        return f"<SCEnvMetaEntry [{', '.join(out)}]>"



from enum import IntFlag
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..strkey import StrKey
from ..utils import raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["TrustLineFlags", "SetTrustLineFlags"]


class TrustLineFlags(IntFlag):
    """Indicates which flags to set. For details about the flags,
    please refer to the `CAP-0035 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0035.md>`_.

    - **AUTHORIZED_FLAG**: issuer has authorized account to perform transactions with its credit

    - **AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG**: issuer has authorized account to maintain and reduce liabilities for its credit

    - **TRUSTLINE_CLAWBACK_ENABLED_FLAG**: issuer has specified that it may clawback its credit, and that claimable balances created with its credit may also be clawed back
    """

    AUTHORIZED_FLAG = 1
    AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG = 2
    TRUSTLINE_CLAWBACK_ENABLED_FLAG = 4


class SetTrustLineFlags(Operation):
    """The :class:`SetTrustLineFlags` object, which represents a SetTrustLineFlags operation on
    Stellar's network.

    Updates the flags of an existing trust line.
    This is called by the issuer of the related asset.

    Threshold: Low

    See `Set Trustline Flags <https://developers.stellar.org/docs/start/list-of-operations/#set-trustline-flags>`_ for more information.

    :param trustor: The account whose trustline this is.
    :param asset: The asset on the trustline.
    :param clear_flags: The flags to clear.
    :param set_flags: The flags to set.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.SET_TRUST_LINE_FLAGS
    )

    def __init__(
        self,
        trustor: str,
        asset: Asset,
        clear_flags: TrustLineFlags = None,
        set_flags: TrustLineFlags = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        super().__init__(source)
        self.trustor: str = trustor
        self.asset: Asset = asset
        self.clear_flags: Optional[TrustLineFlags] = clear_flags
        self.set_flags: Optional[TrustLineFlags] = set_flags
        raise_if_not_valid_ed25519_public_key(self.trustor, "trustor")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        trustor = Keypair.from_public_key(self.trustor).xdr_account_id()
        clear_flags = 0 if self.clear_flags is None else self.clear_flags.value
        set_flags = 0 if self.set_flags is None else self.set_flags.value
        set_trust_line_flags_op = stellar_xdr.SetTrustLineFlagsOp(
            trustor,
            self.asset.to_xdr_object(),
            stellar_xdr.Uint32(clear_flags),
            stellar_xdr.Uint32(set_flags),
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            set_trust_line_flags_op=set_trust_line_flags_op,
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "SetTrustLineFlags":
        """Creates a :class:`SetTrustLineFlags` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.set_trust_line_flags_op is not None
        assert (
            xdr_object.body.set_trust_line_flags_op.trustor.account_id.ed25519
            is not None
        )
        trustor = StrKey.encode_ed25519_public_key(
            xdr_object.body.set_trust_line_flags_op.trustor.account_id.ed25519.uint256
        )
        asset = Asset.from_xdr_object(xdr_object.body.set_trust_line_flags_op.asset)
        clear_flags_xdr = xdr_object.body.set_trust_line_flags_op.clear_flags.uint32
        set_flags_xdr = xdr_object.body.set_trust_line_flags_op.set_flags.uint32
        clear_flags = None if clear_flags_xdr == 0 else TrustLineFlags(clear_flags_xdr)
        set_flags = None if set_flags_xdr == 0 else TrustLineFlags(set_flags_xdr)

        op = cls(
            trustor=trustor,
            asset=asset,
            clear_flags=clear_flags,
            set_flags=set_flags,
            source=source,
        )
        return op

    def __str__(self):
        return (
            f"<SetTrustLineFlags [trustor={self.trustor}, asset={self.asset}, "
            f"clear_flags={self.clear_flags}, set_flags={self.set_flags}, source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .asset_code12 import AssetCode12

__all__ = ["AlphaNum12"]


class AlphaNum12:
    """
    XDR Source Code::

        struct AlphaNum12
        {
            AssetCode12 assetCode;
            AccountID issuer;
        };
    """

    def __init__(
        self,
        asset_code: AssetCode12,
        issuer: AccountID,
    ) -> None:
        self.asset_code = asset_code
        self.issuer = issuer

    def pack(self, packer: Packer) -> None:
        self.asset_code.pack(packer)
        self.issuer.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AlphaNum12:
        asset_code = AssetCode12.unpack(unpacker)
        issuer = AccountID.unpack(unpacker)
        return cls(
            asset_code=asset_code,
            issuer=issuer,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AlphaNum12:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AlphaNum12:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.asset_code,
                self.issuer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.asset_code == other.asset_code and self.issuer == other.issuer

    def __str__(self):
        out = [
            f"asset_code={self.asset_code}",
            f"issuer={self.issuer}",
        ]
        return f"<AlphaNum12 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .curve25519_public import Curve25519Public
from .signature import Signature
from .uint64 import Uint64

__all__ = ["AuthCert"]


class AuthCert:
    """
    XDR Source Code::

        struct AuthCert
        {
            Curve25519Public pubkey;
            uint64 expiration;
            Signature sig;
        };
    """

    def __init__(
        self,
        pubkey: Curve25519Public,
        expiration: Uint64,
        sig: Signature,
    ) -> None:
        self.pubkey = pubkey
        self.expiration = expiration
        self.sig = sig

    def pack(self, packer: Packer) -> None:
        self.pubkey.pack(packer)
        self.expiration.pack(packer)
        self.sig.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AuthCert:
        pubkey = Curve25519Public.unpack(unpacker)
        expiration = Uint64.unpack(unpacker)
        sig = Signature.unpack(unpacker)
        return cls(
            pubkey=pubkey,
            expiration=expiration,
            sig=sig,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AuthCert:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AuthCert:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.pubkey,
                self.expiration,
                self.sig,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.pubkey == other.pubkey
            and self.expiration == other.expiration
            and self.sig == other.sig
        )

    def __str__(self):
        out = [
            f"pubkey={self.pubkey}",
            f"expiration={self.expiration}",
            f"sig={self.sig}",
        ]
        return f"<AuthCert [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .base import Integer
from .tx_set_component import TxSetComponent

__all__ = ["TransactionPhase"]


class TransactionPhase:
    """
    XDR Source Code::

        union TransactionPhase switch (int v)
        {
        case 0:
            TxSetComponent v0Components<>;
        };
    """

    def __init__(
        self,
        v: int,
        v0_components: List[TxSetComponent] = None,
    ) -> None:
        _expect_max_length = 4294967295
        if v0_components and len(v0_components) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `v0_components` should be {_expect_max_length}, but got {len(v0_components)}."
            )
        self.v = v
        self.v0_components = v0_components

    def pack(self, packer: Packer) -> None:
        Integer(self.v).pack(packer)
        if self.v == 0:
            if self.v0_components is None:
                raise ValueError("v0_components should not be None.")
            packer.pack_uint(len(self.v0_components))
            for v0_components_item in self.v0_components:
                v0_components_item.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionPhase:
        v = Integer.unpack(unpacker)
        if v == 0:
            length = unpacker.unpack_uint()
            v0_components = []
            for _ in range(length):
                v0_components.append(TxSetComponent.unpack(unpacker))
            return cls(v=v, v0_components=v0_components)
        return cls(v=v)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionPhase:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionPhase:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.v,
                self.v0_components,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.v == other.v and self.v0_components == other.v0_components

    def __str__(self):
        out = []
        out.append(f"v={self.v}")
        out.append(
            f"v0_components={self.v0_components}"
        ) if self.v0_components is not None else None
        return f"<TransactionPhase [{', '.join(out)}]>"



from ...call_builder.base import BaseRootCallBuilder
from ...call_builder.call_builder_sync.base_call_builder import BaseCallBuilder
from ...client.base_sync_client import BaseSyncClient

__all__ = ["RootCallBuilder"]


class RootCallBuilder(BaseCallBuilder, BaseRootCallBuilder):
    """Creates a new :class:`RootCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.Server.root`.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(self, horizon_url: str, client: BaseSyncClient) -> None:
        super().__init__(horizon_url=horizon_url, client=client)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_spec_entry_kind import SCSpecEntryKind
from .sc_spec_function_v0 import SCSpecFunctionV0
from .sc_spec_udt_enum_v0 import SCSpecUDTEnumV0
from .sc_spec_udt_error_enum_v0 import SCSpecUDTErrorEnumV0
from .sc_spec_udt_struct_v0 import SCSpecUDTStructV0
from .sc_spec_udt_union_v0 import SCSpecUDTUnionV0

__all__ = ["SCSpecEntry"]


class SCSpecEntry:
    """
    XDR Source Code::

        union SCSpecEntry switch (SCSpecEntryKind kind)
        {
        case SC_SPEC_ENTRY_FUNCTION_V0:
            SCSpecFunctionV0 functionV0;
        case SC_SPEC_ENTRY_UDT_STRUCT_V0:
            SCSpecUDTStructV0 udtStructV0;
        case SC_SPEC_ENTRY_UDT_UNION_V0:
            SCSpecUDTUnionV0 udtUnionV0;
        case SC_SPEC_ENTRY_UDT_ENUM_V0:
            SCSpecUDTEnumV0 udtEnumV0;
        case SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            SCSpecUDTErrorEnumV0 udtErrorEnumV0;
        };
    """

    def __init__(
        self,
        kind: SCSpecEntryKind,
        function_v0: SCSpecFunctionV0 = None,
        udt_struct_v0: SCSpecUDTStructV0 = None,
        udt_union_v0: SCSpecUDTUnionV0 = None,
        udt_enum_v0: SCSpecUDTEnumV0 = None,
        udt_error_enum_v0: SCSpecUDTErrorEnumV0 = None,
    ) -> None:
        self.kind = kind
        self.function_v0 = function_v0
        self.udt_struct_v0 = udt_struct_v0
        self.udt_union_v0 = udt_union_v0
        self.udt_enum_v0 = udt_enum_v0
        self.udt_error_enum_v0 = udt_error_enum_v0

    def pack(self, packer: Packer) -> None:
        self.kind.pack(packer)
        if self.kind == SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0:
            if self.function_v0 is None:
                raise ValueError("function_v0 should not be None.")
            self.function_v0.pack(packer)
            return
        if self.kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            if self.udt_struct_v0 is None:
                raise ValueError("udt_struct_v0 should not be None.")
            self.udt_struct_v0.pack(packer)
            return
        if self.kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            if self.udt_union_v0 is None:
                raise ValueError("udt_union_v0 should not be None.")
            self.udt_union_v0.pack(packer)
            return
        if self.kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            if self.udt_enum_v0 is None:
                raise ValueError("udt_enum_v0 should not be None.")
            self.udt_enum_v0.pack(packer)
            return
        if self.kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            if self.udt_error_enum_v0 is None:
                raise ValueError("udt_error_enum_v0 should not be None.")
            self.udt_error_enum_v0.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecEntry:
        kind = SCSpecEntryKind.unpack(unpacker)
        if kind == SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0:
            function_v0 = SCSpecFunctionV0.unpack(unpacker)
            return cls(kind=kind, function_v0=function_v0)
        if kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            udt_struct_v0 = SCSpecUDTStructV0.unpack(unpacker)
            return cls(kind=kind, udt_struct_v0=udt_struct_v0)
        if kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            udt_union_v0 = SCSpecUDTUnionV0.unpack(unpacker)
            return cls(kind=kind, udt_union_v0=udt_union_v0)
        if kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            udt_enum_v0 = SCSpecUDTEnumV0.unpack(unpacker)
            return cls(kind=kind, udt_enum_v0=udt_enum_v0)
        if kind == SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            udt_error_enum_v0 = SCSpecUDTErrorEnumV0.unpack(unpacker)
            return cls(kind=kind, udt_error_enum_v0=udt_error_enum_v0)
        return cls(kind=kind)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.kind,
                self.function_v0,
                self.udt_struct_v0,
                self.udt_union_v0,
                self.udt_enum_v0,
                self.udt_error_enum_v0,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.kind == other.kind
            and self.function_v0 == other.function_v0
            and self.udt_struct_v0 == other.udt_struct_v0
            and self.udt_union_v0 == other.udt_union_v0
            and self.udt_enum_v0 == other.udt_enum_v0
            and self.udt_error_enum_v0 == other.udt_error_enum_v0
        )

    def __str__(self):
        out = []
        out.append(f"kind={self.kind}")
        out.append(
            f"function_v0={self.function_v0}"
        ) if self.function_v0 is not None else None
        out.append(
            f"udt_struct_v0={self.udt_struct_v0}"
        ) if self.udt_struct_v0 is not None else None
        out.append(
            f"udt_union_v0={self.udt_union_v0}"
        ) if self.udt_union_v0 is not None else None
        out.append(
            f"udt_enum_v0={self.udt_enum_v0}"
        ) if self.udt_enum_v0 is not None else None
        out.append(
            f"udt_error_enum_v0={self.udt_error_enum_v0}"
        ) if self.udt_error_enum_v0 is not None else None
        return f"<SCSpecEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .tx_set_component_txs_maybe_discounted_fee import (
    TxSetComponentTxsMaybeDiscountedFee,
)
from .tx_set_component_type import TxSetComponentType

__all__ = ["TxSetComponent"]


class TxSetComponent:
    """
    XDR Source Code::

        union TxSetComponent switch (TxSetComponentType type)
        {
        case TXSET_COMP_TXS_MAYBE_DISCOUNTED_FEE:
          struct
          {
            int64* baseFee;
            TransactionEnvelope txs<>;
          } txsMaybeDiscountedFee;
        };
    """

    def __init__(
        self,
        type: TxSetComponentType,
        txs_maybe_discounted_fee: TxSetComponentTxsMaybeDiscountedFee = None,
    ) -> None:
        self.type = type
        self.txs_maybe_discounted_fee = txs_maybe_discounted_fee

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == TxSetComponentType.TXSET_COMP_TXS_MAYBE_DISCOUNTED_FEE:
            if self.txs_maybe_discounted_fee is None:
                raise ValueError("txs_maybe_discounted_fee should not be None.")
            self.txs_maybe_discounted_fee.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TxSetComponent:
        type = TxSetComponentType.unpack(unpacker)
        if type == TxSetComponentType.TXSET_COMP_TXS_MAYBE_DISCOUNTED_FEE:
            txs_maybe_discounted_fee = TxSetComponentTxsMaybeDiscountedFee.unpack(
                unpacker
            )
            return cls(type=type, txs_maybe_discounted_fee=txs_maybe_discounted_fee)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TxSetComponent:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TxSetComponent:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.txs_maybe_discounted_fee,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.txs_maybe_discounted_fee == other.txs_maybe_discounted_fee
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"txs_maybe_discounted_fee={self.txs_maybe_discounted_fee}"
        ) if self.txs_maybe_discounted_fee is not None else None
        return f"<TxSetComponent [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .scp_envelope import SCPEnvelope
from .uint32 import Uint32

__all__ = ["LedgerSCPMessages"]


class LedgerSCPMessages:
    """
    XDR Source Code::

        struct LedgerSCPMessages
        {
            uint32 ledgerSeq;
            SCPEnvelope messages<>;
        };
    """

    def __init__(
        self,
        ledger_seq: Uint32,
        messages: List[SCPEnvelope],
    ) -> None:
        _expect_max_length = 4294967295
        if messages and len(messages) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `messages` should be {_expect_max_length}, but got {len(messages)}."
            )
        self.ledger_seq = ledger_seq
        self.messages = messages

    def pack(self, packer: Packer) -> None:
        self.ledger_seq.pack(packer)
        packer.pack_uint(len(self.messages))
        for messages_item in self.messages:
            messages_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerSCPMessages:
        ledger_seq = Uint32.unpack(unpacker)
        length = unpacker.unpack_uint()
        messages = []
        for _ in range(length):
            messages.append(SCPEnvelope.unpack(unpacker))
        return cls(
            ledger_seq=ledger_seq,
            messages=messages,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerSCPMessages:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerSCPMessages:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_seq,
                self.messages,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ledger_seq == other.ledger_seq and self.messages == other.messages

    def __str__(self):
        out = [
            f"ledger_seq={self.ledger_seq}",
            f"messages={self.messages}",
        ]
        return f"<LedgerSCPMessages [{', '.join(out)}]>"



from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..strkey import StrKey
from ..utils import raise_if_not_valid_amount, raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["CreateAccount"]


class CreateAccount(Operation):
    """The :class:`CreateAccount` object, which represents a Create Account
    operation on Stellar's network.

    This operation creates and funds a new account with the specified starting
    balance.

    Threshold: Medium

    See `Create Account <https://developers.stellar.org/docs/start/list-of-operations/#create-account>`_ for more information.

    :param destination: Destination account ID to create an account for.
    :param starting_balance: Amount in XLM the account should be
        funded for. Must be greater than the `reserve balance amount
        <https://developers.stellar.org/docs/glossary/minimum-balance/>`_.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.CREATE_ACCOUNT
    )

    def __init__(
        self,
        destination: str,
        starting_balance: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        self.destination: str = destination
        self.starting_balance: str = str(starting_balance)
        raise_if_not_valid_ed25519_public_key(self.destination, "destination")
        raise_if_not_valid_amount(self.starting_balance, "starting_balance")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        destination = Keypair.from_public_key(self.destination).xdr_account_id()
        starting_balance = stellar_xdr.Int64(
            Operation.to_xdr_amount(self.starting_balance)
        )
        create_account_op = stellar_xdr.CreateAccountOp(destination, starting_balance)
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, create_account_op=create_account_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "CreateAccount":
        """Creates a :class:`CreateAccount` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.create_account_op is not None
        assert (
            xdr_object.body.create_account_op.destination.account_id.ed25519 is not None
        )
        destination = StrKey.encode_ed25519_public_key(
            xdr_object.body.create_account_op.destination.account_id.ed25519.uint256
        )
        starting_balance = Operation.from_xdr_amount(
            xdr_object.body.create_account_op.starting_balance.int64
        )
        op = cls(
            source=source, destination=destination, starting_balance=starting_balance
        )
        return op

    def __str__(self):
        return (
            f"<CreateAccount [destination={self.destination}, "
            f"starting_balance={self.starting_balance}, "
            f"source={self.source}]>"
        )



import copy
import random
from typing import Callable, Optional, Union

from . import scval
from . import xdr as stellar_xdr
from .address import Address
from .exceptions import BadSignatureError
from .keypair import Keypair
from .network import Network
from .utils import sha256

__all__ = ["authorize_entry", "authorize_invocation"]


def authorize_entry(
    entry: Union[stellar_xdr.SorobanAuthorizationEntry, str],
    signer: Union[Keypair, Callable[[stellar_xdr.HashIDPreimage], bytes]],
    valid_until_ledger_sequence: int,
    network_passphrase: str,
) -> stellar_xdr.SorobanAuthorizationEntry:
    """Actually authorizes an existing authorization entry using the given the
    credentials and expiration details, returning a signed copy.

    This "fills out" the authorization entry with a signature, indicating to the
    :class:`stellar_sdk.InvokeHostFunction` it's attached to that:

    * a particular identity (i.e. signing :class:`stellar_sdk.Keypair` or other signer)
    * approving the execution of an invocation tree (i.e. a
        simulation-acquired :class:`stellar_xdr.SorobanAuthorizedInvocation` or otherwise built)
    * on a particular network (uniquely identified by its passphrase, see :class:`stellar_sdk.Network`)
    * until a particular ledger sequence is reached.

    :param entry: an unsigned Soroban authorization entry.
    :param signer: either a :class:`Keypair` or a function which takes a payload
        (a :class:`stellar_xdr.HashIDPreimage` instance) input and returns a bytes signature,
        the signing key should correspond to the address in the `entry`.
    :param valid_until_ledger_sequence: the (exclusive) future ledger sequence number until which
        this authorization entry should be valid (if `currentLedgerSeq==validUntil`, this is expired)
    :param network_passphrase: the network passphrase is incorporated into the signature (see :class:`stellar_sdk.Network` for options)
    :return: a signed Soroban authorization entry.
    """

    if isinstance(entry, str):
        entry = stellar_xdr.SorobanAuthorizationEntry.from_xdr(entry)
    else:
        entry = copy.deepcopy(entry)

    if (
        entry.credentials.type
        != stellar_xdr.SorobanCredentialsType.SOROBAN_CREDENTIALS_ADDRESS
    ):
        return entry

    assert entry.credentials.address is not None
    addr_auth = entry.credentials.address
    addr_auth.signature_expiration_ledger = stellar_xdr.Uint32(
        valid_until_ledger_sequence
    )

    network_id = Network(network_passphrase).network_id()
    preimage = stellar_xdr.HashIDPreimage(
        type=stellar_xdr.EnvelopeType.ENVELOPE_TYPE_SOROBAN_AUTHORIZATION,
        soroban_authorization=stellar_xdr.HashIDPreimageSorobanAuthorization(
            network_id=stellar_xdr.Hash(network_id),
            nonce=addr_auth.nonce,
            signature_expiration_ledger=addr_auth.signature_expiration_ledger,
            invocation=entry.root_invocation,
        ),
    )
    payload = sha256(preimage.to_xdr_bytes())
    if isinstance(signer, Keypair):
        signature = signer.sign(payload)
    else:
        signature = signer(preimage)

    public_key = Address.from_xdr_sc_address(addr_auth.address).key
    try:
        Keypair.from_raw_ed25519_public_key(public_key).verify(payload, signature)
    except BadSignatureError as e:
        raise ValueError("signature doesn't match payload.") from e

    # This structure is defined here:
    # https://soroban.stellar.org/docs/fundamentals-and-concepts/invoking-contracts-with-transactions#stellar-account-signatures
    addr_auth.signature = scval.to_vec(
        [
            scval.to_map(
                {
                    scval.to_symbol("public_key"): scval.to_bytes(public_key),
                    scval.to_symbol("signature"): scval.to_bytes(signature),
                }
            )
        ]
    )
    return entry


def authorize_invocation(
    signer: Union[Keypair, Callable[[stellar_xdr.HashIDPreimage], bytes]],
    public_key: Optional[str],
    valid_until_ledger_sequence: int,
    invocation: stellar_xdr.SorobanAuthorizedInvocation,
    network_passphrase: str,
):
    """This builds an entry from scratch, allowing you to express authorization as a function of:

    * a particular identity (i.e. signing :class:`stellar_sdk.Keypair` or other signer)
    * approving the execution of an invocation tree (i.e. a
        simulation-acquired :class:`stellar_xdr.SorobanAuthorizedInvocation` or otherwise built)
    * on a particular network (uniquely identified by its passphrase, see :class:`stellar_sdk.Network`)
    * until a particular ledger sequence is reached.

    This is in contrast to :func:`authorize_entry`, which signs an existing entry "in place".

    :param signer: either a :class:`Keypair` or a function which takes a payload
        (a :class:`stellar_xdr.HashIDPreimage` instance) input and returns a bytes signature,
        the signing key should correspond to the address in the `entry`.
    :param public_key: the public identity of the signer (when providing a :class:`Keypair` to `signer`,
        this can be omitted, as it just uses the public key of the keypair)
    :param valid_until_ledger_sequence: the (exclusive) future ledger sequence number until which
        this authorization entry should be valid (if `currentLedgerSeq==validUntil`, this is expired)
    :param invocation: invocation the invocation tree that we're authorizing (likely, this comes from transaction simulation)
    :param network_passphrase: the network passphrase is incorporated into the signature (see :class:`stellar_sdk.Network` for options)
    :return: a signed Soroban authorization entry.
    """
    nonce = random.randint(-(2**63), 2**63 - 1)
    pk = public_key
    if not pk and isinstance(signer, Keypair):
        pk = signer.public_key

    if not pk:
        raise ValueError("`public_key` parameter is required.")

    entry = stellar_xdr.SorobanAuthorizationEntry(
        root_invocation=invocation,
        credentials=stellar_xdr.SorobanCredentials(
            type=stellar_xdr.SorobanCredentialsType.SOROBAN_CREDENTIALS_ADDRESS,
            address=stellar_xdr.SorobanAddressCredentials(
                address=Address(pk).to_xdr_sc_address(),
                nonce=stellar_xdr.Int64(nonce),
                signature_expiration_ledger=stellar_xdr.Uint32(0),
                signature=stellar_xdr.SCVal(type=stellar_xdr.SCValType.SCV_VOID),
            ),
        ),
    )
    return authorize_entry(
        entry, signer, valid_until_ledger_sequence, network_passphrase
    )



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["ExtendFootprintTTL"]


class ExtendFootprintTTL(Operation):
    """The :class:`ExtendFootprintTTL` object, which represents a ExtendFootprintTTL
    operation on Stellar's network.

    Threshold: Medium

    See `ExtendFootprintTTLOp <https://soroban.stellar.org/docs/fundamentals-and-concepts/state-expiration#ExtendFootprintTTLop>`_.

    :param extend_to: The number of ledgers past the LCL (last closed ledger)
        by which to extend the validity of the ledger keys in this transaction.
    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.EXTEND_FOOTPRINT_TTL
    )

    def __init__(
        self, extend_to: int, source: Optional[Union[MuxedAccount, str]] = None
    ) -> None:
        super().__init__(source)
        if extend_to < 0 or extend_to > 2**32 - 1:
            raise ValueError(
                f"`extend_to` value must be between 0 and 2**32-1, got {extend_to}"
            )

        self.extend_to: int = extend_to

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        op = stellar_xdr.ExtendFootprintTTLOp(
            ext=stellar_xdr.ExtensionPoint(0),
            extend_to=stellar_xdr.Uint32(self.extend_to),
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, extend_footprint_ttl_op=op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "ExtendFootprintTTL":
        """Creates a :class:`ExtendFootprintTTL` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.extend_footprint_ttl_op is not None
        extend_to = xdr_object.body.extend_footprint_ttl_op.extend_to.uint32
        op = cls(source=source, extend_to=extend_to)
        return op

    def __str__(self):
        return (
            f"<ExtendFootprintTTL [extend_to={self.extend_to}, source={self.source}]>"
        )



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["RestoreFootprint"]


class RestoreFootprint(Operation):
    """The :class:`RestoreFootprint` object, which represents a RestoreFootprint
    operation on Stellar's network.

    Threshold: Medium

    See `RestoreFootprintOp <https://soroban.stellar.org/docs/fundamentals-and-concepts/state-expiration#restorefootprintop>`_.

    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.RESTORE_FOOTPRINT
    )

    def __init__(self, source: Optional[Union[MuxedAccount, str]] = None) -> None:
        super().__init__(source)

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        op = stellar_xdr.RestoreFootprintOp(
            ext=stellar_xdr.ExtensionPoint(0),
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, restore_footprint_op=op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "RestoreFootprint":
        """Creates a :class:`RestoreFootprint` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        op = cls(source=source)
        return op

    def __str__(self):
        return f"<RestoreFootprint [source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque
from .create_contract_args import CreateContractArgs
from .host_function_type import HostFunctionType
from .invoke_contract_args import InvokeContractArgs

__all__ = ["HostFunction"]


class HostFunction:
    """
    XDR Source Code::

        union HostFunction switch (HostFunctionType type)
        {
        case HOST_FUNCTION_TYPE_INVOKE_CONTRACT:
            InvokeContractArgs invokeContract;
        case HOST_FUNCTION_TYPE_CREATE_CONTRACT:
            CreateContractArgs createContract;
        case HOST_FUNCTION_TYPE_UPLOAD_CONTRACT_WASM:
            opaque wasm<>;
        };
    """

    def __init__(
        self,
        type: HostFunctionType,
        invoke_contract: InvokeContractArgs = None,
        create_contract: CreateContractArgs = None,
        wasm: bytes = None,
    ) -> None:
        self.type = type
        self.invoke_contract = invoke_contract
        self.create_contract = create_contract
        self.wasm = wasm

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == HostFunctionType.HOST_FUNCTION_TYPE_INVOKE_CONTRACT:
            if self.invoke_contract is None:
                raise ValueError("invoke_contract should not be None.")
            self.invoke_contract.pack(packer)
            return
        if self.type == HostFunctionType.HOST_FUNCTION_TYPE_CREATE_CONTRACT:
            if self.create_contract is None:
                raise ValueError("create_contract should not be None.")
            self.create_contract.pack(packer)
            return
        if self.type == HostFunctionType.HOST_FUNCTION_TYPE_UPLOAD_CONTRACT_WASM:
            if self.wasm is None:
                raise ValueError("wasm should not be None.")
            Opaque(self.wasm, 4294967295, False).pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HostFunction:
        type = HostFunctionType.unpack(unpacker)
        if type == HostFunctionType.HOST_FUNCTION_TYPE_INVOKE_CONTRACT:
            invoke_contract = InvokeContractArgs.unpack(unpacker)
            return cls(type=type, invoke_contract=invoke_contract)
        if type == HostFunctionType.HOST_FUNCTION_TYPE_CREATE_CONTRACT:
            create_contract = CreateContractArgs.unpack(unpacker)
            return cls(type=type, create_contract=create_contract)
        if type == HostFunctionType.HOST_FUNCTION_TYPE_UPLOAD_CONTRACT_WASM:
            wasm = Opaque.unpack(unpacker, 4294967295, False)
            return cls(type=type, wasm=wasm)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HostFunction:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HostFunction:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.invoke_contract,
                self.create_contract,
                self.wasm,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.invoke_contract == other.invoke_contract
            and self.create_contract == other.create_contract
            and self.wasm == other.wasm
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"invoke_contract={self.invoke_contract}"
        ) if self.invoke_contract is not None else None
        out.append(
            f"create_contract={self.create_contract}"
        ) if self.create_contract is not None else None
        out.append(f"wasm={self.wasm}") if self.wasm is not None else None
        return f"<HostFunction [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .survey_message_response_type import SurveyMessageResponseType
from .topology_response_body_v0 import TopologyResponseBodyV0
from .topology_response_body_v1 import TopologyResponseBodyV1

__all__ = ["SurveyResponseBody"]


class SurveyResponseBody:
    """
    XDR Source Code::

        union SurveyResponseBody switch (SurveyMessageResponseType type)
        {
        case SURVEY_TOPOLOGY_RESPONSE_V0:
            TopologyResponseBodyV0 topologyResponseBodyV0;
        case SURVEY_TOPOLOGY_RESPONSE_V1:
            TopologyResponseBodyV1 topologyResponseBodyV1;
        };
    """

    def __init__(
        self,
        type: SurveyMessageResponseType,
        topology_response_body_v0: TopologyResponseBodyV0 = None,
        topology_response_body_v1: TopologyResponseBodyV1 = None,
    ) -> None:
        self.type = type
        self.topology_response_body_v0 = topology_response_body_v0
        self.topology_response_body_v1 = topology_response_body_v1

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SurveyMessageResponseType.SURVEY_TOPOLOGY_RESPONSE_V0:
            if self.topology_response_body_v0 is None:
                raise ValueError("topology_response_body_v0 should not be None.")
            self.topology_response_body_v0.pack(packer)
            return
        if self.type == SurveyMessageResponseType.SURVEY_TOPOLOGY_RESPONSE_V1:
            if self.topology_response_body_v1 is None:
                raise ValueError("topology_response_body_v1 should not be None.")
            self.topology_response_body_v1.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SurveyResponseBody:
        type = SurveyMessageResponseType.unpack(unpacker)
        if type == SurveyMessageResponseType.SURVEY_TOPOLOGY_RESPONSE_V0:
            topology_response_body_v0 = TopologyResponseBodyV0.unpack(unpacker)
            return cls(type=type, topology_response_body_v0=topology_response_body_v0)
        if type == SurveyMessageResponseType.SURVEY_TOPOLOGY_RESPONSE_V1:
            topology_response_body_v1 = TopologyResponseBodyV1.unpack(unpacker)
            return cls(type=type, topology_response_body_v1=topology_response_body_v1)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SurveyResponseBody:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SurveyResponseBody:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.topology_response_body_v0,
                self.topology_response_body_v1,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.topology_response_body_v0 == other.topology_response_body_v0
            and self.topology_response_body_v1 == other.topology_response_body_v1
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"topology_response_body_v0={self.topology_response_body_v0}"
        ) if self.topology_response_body_v0 is not None else None
        out.append(
            f"topology_response_body_v1={self.topology_response_body_v1}"
        ) if self.topology_response_body_v1 is not None else None
        return f"<SurveyResponseBody [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["LiquidityPoolWithdrawResultCode"]


class LiquidityPoolWithdrawResultCode(IntEnum):
    """
    XDR Source Code::

        enum LiquidityPoolWithdrawResultCode
        {
            // codes considered as "success" for the operation
            LIQUIDITY_POOL_WITHDRAW_SUCCESS = 0,

            // codes considered as "failure" for the operation
            LIQUIDITY_POOL_WITHDRAW_MALFORMED = -1,    // bad input
            LIQUIDITY_POOL_WITHDRAW_NO_TRUST = -2,     // no trust line for one of the
                                                       // assets
            LIQUIDITY_POOL_WITHDRAW_UNDERFUNDED = -3,  // not enough balance of the
                                                       // pool share
            LIQUIDITY_POOL_WITHDRAW_LINE_FULL = -4,    // would go above limit for one
                                                       // of the assets
            LIQUIDITY_POOL_WITHDRAW_UNDER_MINIMUM = -5 // didn't withdraw enough
        };
    """

    LIQUIDITY_POOL_WITHDRAW_SUCCESS = 0
    LIQUIDITY_POOL_WITHDRAW_MALFORMED = -1
    LIQUIDITY_POOL_WITHDRAW_NO_TRUST = -2
    LIQUIDITY_POOL_WITHDRAW_UNDERFUNDED = -3
    LIQUIDITY_POOL_WITHDRAW_LINE_FULL = -4
    LIQUIDITY_POOL_WITHDRAW_UNDER_MINIMUM = -5

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolWithdrawResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolWithdrawResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolWithdrawResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from ...call_builder.base import BaseClaimableBalancesCallBuilder
from ...call_builder.call_builder_sync.base_call_builder import BaseCallBuilder
from ...client.base_sync_client import BaseSyncClient

__all__ = ["ClaimableBalancesCallBuilder"]


class ClaimableBalancesCallBuilder(BaseCallBuilder, BaseClaimableBalancesCallBuilder):
    """Creates a new :class:`ClaimableBalancesCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.Server.claimable_balance`.

    See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(self, horizon_url, client: BaseSyncClient) -> None:
        super().__init__(horizon_url=horizon_url, client=client)
